-- Wall size in pixels
local WALL = 16

-- Window Size X and Y
local WSX, WSY = 1024, 768

-- Scrolling speed (px/s)
local SCROLL_SPEED = 120

-- Join probability when groups are equal
local JOIN_PROB = 0.5

-- Wall adding probability
local WALL_ADD_PROB = 0.5

-- Color for passages
local PASSAGE_COLOR = {250, 240, 220}

local lg = love.graphics

-- maxX and maxY are in cells
local maxX = math.floor(WSX / WALL) + 3
local maxY = math.floor((WSY / WALL - 1) / 2) * 2 + 1

local row, counts, passage_counts, pool, perm
local canvas
local canvasX, canvasY = 0, 0
local brick
local batch

local genX = 0
local rnd
do
  require 'socket'
  local t = socket.gettime()
  rnd = love.math.newRandomGenerator(math.floor(t), math.floor((t % 1) * 0x80000000))
end


local function Generate()
  -- The original Eller's algorithm is designed to work row by row,
  -- and we've followed a naming convention here that refers to rows,
  -- even if the output is in columns.

  -- Clear the area to paint
  lg.setColor(PASSAGE_COLOR)
  lg.rectangle("fill", genX, 0, 2 * WALL, maxX * WALL)

  lg.setColor(255, 255, 255)

  local ncells = (maxY - 1) / 2

  batch:clear()

  -- Draw top and bottom bricks
  batch:add(genX, 0)
  batch:add(genX, (maxY - 1) * WALL)

  -- Add walls between cells
  for i = 1, ncells - 1 do
    local add_wall = true
    if row[i] ~= row[i + 1] then

      -- Different sets. Join them?
      if rnd:random() < JOIN_PROB then
        -- Yup
        local set = row[i]
        local tgt = row[i + 1]

        -- Alas, merging the groups is O(n^2). Pullen suggests
        -- doubly linked lists. His implementation seems to rely
        -- on the lists always being sorted, which is an
        -- adventurous assumption. Our n is going to be small
        -- anyway, so O(n^2) is OK.
        for j = 1, ncells do
          if row[j] == tgt then
        
            row[j] = set -- overwrite tgt with set
            counts[set] = counts[set] + 1 -- update counter
            counts[tgt] = counts[tgt] - 1 -- update counter
            if counts[tgt] == 0 then
              pool[#pool + 1] = tgt -- return to pool
            end

          end -- same set check
        end -- for

        add_wall = false -- in this case, don't add a wall
      end -- if different sets and decided to join them

    end -- if different sets

    if add_wall then
      batch:add(genX, 2 * WALL * i)
    end
  end -- of loop to check pairs of adjacent cells

  -- Advance coordinate
  genX = genX + WALL
  if genX >= maxX * WALL then
    genX = 0
  end

  -- Draw top wall
  batch:add(genX, 0)

  -- Init vertical walls' variables
  for i = 1, ncells do
    passage_counts[i] = counts[i]
    perm[i] = i

    -- The cells above and below always have a wall. Paint the one below
    -- (we've already painted the one on top of all, and the wall below
    --  is the one above the next cell)
    batch:add(genX, 2 * WALL * i)
  end

  --[[--
          Eller's algorithm has a difficult step here: to carve passages down
          at random, but ensuring that every set has at least one passage 
          leading down.

          Our solution is to start with no walls, and visit the cells in a
          random order (by generating a random permutation of cell indexes
          first), keeping account of how many cells per group have walls
          added. When the count is down to 1, don't add a wall.
  --]]--

  -- Generate random permutation using inside-out Fisher-Yates algorithm
  -- http://orden-y-concierto.blogspot.com/2013/08/inside-out-fisher-yates-algorithm.html
  for i = 1, ncells do
    local r = rnd:random(1, i)
    perm[i] = perm[r]
    perm[r] = i
  end

  -- Add walls to the next row
  for i = 1, ncells do
    local idx = perm[i]
    local set = row[idx]
    if passage_counts[set] ~= 1 then
      if rnd:random() < WALL_ADD_PROB then
        passage_counts[set] = passage_counts[set] - 1
        batch:add(genX, WALL * (idx * 2 - 1))

        -- Assign a new set to it
        -- We're going to overwrite it, so update groups
        counts[set] = counts[set] - 1
        -- Not needed because we know there's at least one
        --if counts[set] == 0 then
        --  pool[#pool] = set
        --end

        set = table.remove(pool)
        row[idx] = set
        counts[set] = 1
      end
    end
  end

  -- Advance coordinate
  genX = genX + WALL
  if genX >= maxX * WALL then
    genX = 0
  end

  lg.draw(batch)
end


function love.load(args)
  -- Init the arrays for the algorithm (preallocation)
  local ncells = (maxY - 1) / 2
  row = {}
  counts = {}
  passage_counts = {}
  pool = {}
  perm = {}
  for i = 1, ncells do
    row[i] = i
    counts[i] = 1
    passage_counts[i] = 1
    perm[i] = i
  end

  love.window.setMode(WSX, WSY)

  -- Draw brick image
  brick = lg.newCanvas(WALL, WALL)
  lg.setCanvas(brick)
  lg.setColor(160, 0, 0)
  lg.rectangle("fill", 0, 0, WALL, WALL)
  lg.setColor(255, 255, 255)
  for i = 3, WALL-1, 4 do
    lg.rectangle("fill", 0, i, WALL, 1)
  end
  for i = 0, WALL-1, 8 do
    lg.rectangle("fill", WALL/8, i, 1, 3)
    lg.rectangle("fill", WALL/8 + WALL/4, i + 4, 1, 3)
    lg.rectangle("fill", WALL/8 + WALL/2, i, 1, 3)
    lg.rectangle("fill", WALL/8 + WALL/4 + WALL/2, i + 4, 1, 3)
  end
  batch = lg.newSpriteBatch(brick, math.ceil(maxX * 1.5), "dynamic")

  canvas = lg.newCanvas(maxX * WALL, maxY * WALL)
  lg.setCanvas(canvas)
  lg.clear(PASSAGE_COLOR)

  for i = 1, math.floor(maxX/2) do Generate() end
end


function love.update(dt)
  local tmp = canvasX

  canvasX = canvasX - math.min(SCROLL_SPEED * dt, 2 * WALL)
  if canvasX < -maxX * WALL then
    canvasX = canvasX + maxX * WALL
  end

  local draw_point = genX + canvasX
  if draw_point < 0 then
    draw_point = draw_point + maxX * WALL
  end
  if draw_point >= 0 and draw_point < WSX then
    Generate()
  end

  lg.setCanvas()
end


function love.draw()
  lg.setColor(255, 255, 255)
  lg.draw(canvas, math.floor(canvasX), canvasY)
  lg.draw(canvas, math.floor(canvasX) + maxX * WALL, canvasY)
  love.window.setTitle(love.timer.getFPS())
  lg.setCanvas(canvas)
end


function love.keypressed(k)
  if k == "escape" then love.event.quit() end
end
